// *#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* AvFifo_VideoRx.cpp *#*#*#*#*#*#*#*#*# (C) 2023 DekTec
//
// Code example AvFifo_VideoRx - SMPTE 2110 video receiver and display using SDL
//
// This code serves as a practical code example of receiving a SMPTE 2110 video stream
// and displaying it in a window using the SDL library with video scaling by 0.5.
// 
// The example covers key concepts such as device configuration, receive FIFO management, 
// processing of received frames, SDL event handling, resource cleanup, and exception 
// handling.
// 
// Notes:
//  - This example uses a fixed IP address, port number, video format, etc.
//    These configuration parameters are defined in the 'Configuration' section below.
//  - The application downscales the video by a factor of 2, reducing the size of the 
//    video content to fit into the display window.
// 

const char* Version{"Jun2023.2"};

// -.-.-.-.-.-.-.-.-.-.-.- Include necessary headers and libraries -.-.-.-.-.-.-.-.-.-.-.-

#include "DTAPI.h"                  // C++ API for DekTec devices.
#include "DTAPI_AvFifo.h"           // DekTec's AvFifo API, part of DTAPI.
#include <SDL.h>                    // Cross-platform library providing low-level access 
                                    // to video, audio and other I/O functions.
#include <chrono>                   // Standard library: date and time utilities.
#include <iostream>                 // Standard library: input/output operations.
#include <thread>                   // Standard library: Thread support.

using namespace std;                // Avoid std::prefix for standard library calls.

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Configuration -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

// If a config value can be changed without changing the application code, it is 
// labeled "CONFIGURABLE".

// DTA-2110 device number: 0=First DTA-2110 in the system, 1=Second one, etc.
constexpr int Dta2110DeviceNo{0};                // CONFIGURABLE

// Fixed IP receive configuration.
const AvFifo::IpPars AvFifo_VideoRx_IpPars
{
    .IpAddr{239, 10, 10, 10},                    // CONFIGURABLE
    .IpVersion{AvFifo::IpProtocolVersion::IPv4}, // CONFIGURABLE
    .Port{5000},                                 // CONFIGURABLE
    // For SMPTE 2110, transport protocol must be set to RTP.
    .TransportProtocol{AvFifo::IpTransportProtocol::Rtp}
};

// Fixed SMPTE 2110 video format.
// We choose format Uyvy422_10b_to_8b, which lets the AvFifo library automatically 
// convert the 10-bit UYVY pixels to 8-bit, for easy processing of the frame data.
// Note: Changing the format requires rewriting this program.
const AvFifo::St2110::RxConfigVideo AvFifo_VideoRx_Config
{
    .Format{AvFifo::St2110::RxFrameFormat::Uyvy422_10b_to_8b}
};

// Fixed incoming frame size (1920x1080).
constexpr int FrameWidth{1920};     // CONFIGURABLE
constexpr int FrameHeight{1080};    // CONFIGURABLE

// The factor by which to scale down the video when displaying in an SDL window.
constexpr int Scale{2};             // CONFIGURABLE

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

// Stores the SDL state and takes care of proper initialization/destruction.
struct SdlContext
{
    SdlContext();                   // Initializes SDL.
    ~SdlContext();                  // Cleans up SDL.
    void Render(AvFifo::Frame*);    // Renders a frame to the Window.
    SDL_Renderer* Renderer{};
    SDL_Texture* Texture{};
    SDL_Window* Window{};
};

// Forward declarations to enable top-down function presentation.
void AvFifo_VideoRx();
void VideoRxLoop(DtDevice&);

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- main -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// The main function prints a welcome message and starts the main AvFifo_VideoRx()
// function, catching exceptions and displaying appropriate error messages.
// Exceptions at this level suggest a critical failure that prevents the program from 
// running. For example, the absence of a DTA-2110 in the system triggers such an
// exception.
//
int main(int argc, char* argv[])
{
    cout << "\n-----------------------------------------------------------------------\n"
         << "DekTec AvFifo API Example - AvFifo_VideoRx - SMPTE 2110 video receiver\n"
         << "Version " << Version << "\n"
         << "-----------------------------------------------------------------------\n\n"
         << "DTA-2110 device number: " << Dta2110DeviceNo << "\n"
         << "Expected IP address   : "
         << (int)AvFifo_VideoRx_IpPars.IpAddr[0] << "."
         << (int)AvFifo_VideoRx_IpPars.IpAddr[1] << "."
         << (int)AvFifo_VideoRx_IpPars.IpAddr[2] << "."
         << (int)AvFifo_VideoRx_IpPars.IpAddr[3] << ", "
         << "port " << AvFifo_VideoRx_IpPars.Port << "\n"
         << "Expected video format : 10-bit packed, 4:2:2 UYVY chroma subsampling, "
         << FrameWidth << "x" << FrameHeight << "\n"
         << "Scale-down factor     : " << Scale << "\n\n"
         << "Closing the Video Window will quit the application.\n";

    // Wrapping the execution of AvFifo_VideoRx() function in a try block 
    // allows us to handle any exceptions that might be thrown during its execution.
    try {AvFifo_VideoRx();}

    // UsageError exceptions indicate programming errors.
    catch (AvFifo::UsageError& e) {cerr << "Usage error: " << e.what() << '\n';}

    // If a standard exception is thrown (e.g. bad_alloc), catch it and print error 
    // message.
    catch (exception& e) {cerr << "Exception: " << e.what() << '\n';}

    // This is a catch-all handler for any exceptions not caught above.
    // If an exception type not caught in the previous handlers is thrown, 
    // it will be caught here and an error message will be displayed.
    catch (...) {cerr << "Caught an unknown exception\n";}

    return 0;
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- AvFifo_VideoRx -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Attaches to the hardware, invokes the main loop, and detaches from the hardware.
//
void AvFifo_VideoRx()
{
    // Attach DtDevice object to the DTA-2110 PCIe card with index 'Dta2110DeviceNo'.
    DtDevice Dta2110{};
    DTAPI_RESULT Res{Dta2110.AttachToType(2110, Dta2110DeviceNo)};
    if (Res != DTAPI_OK)
        throw runtime_error("DTA-2110 attach fails: " + string(DtapiResult2Str(Res)));

    // Check if network port 1 is operational before proceeding.
    // This step is essential for auto-starting applications since the network stack 
    // may not be immediately fully functional after the PC boots.
    Res = Dta2110.IsNetworkCardOperational(1, 
                        AvFifo_VideoRx_IpPars.IpVersion==AvFifo::IpProtocolVersion::IPv4,
                        AvFifo_VideoRx_IpPars.IpVersion==AvFifo::IpProtocolVersion::IPv6);
    if (Res != DTAPI_OK)
    {
        Dta2110.Detach();
        throw runtime_error("Networking error: " + string(DtapiResult2Str(Res)));
    }

    // Network card is operational, proceed to invoke the main loop.
    try {VideoRxLoop(Dta2110);}

    // Catch and report driver errors.
    catch (AvFifo::DriverError& e) {cerr << "Driver error: " << e.what() << '\n';}

    // Catch and report unavailable hardware pipe errors.
    catch (AvFifo::HardwarePipeUnavailable& e) {cerr << e.what() << '\n';}

    // Cleanup: Detach the Dta2110 object from the hardware.
    Dta2110.Detach();
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- VideoRxLoop -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// This function is the core of the example program.
// It handles the receiving of video frames and rendering them in a window.
//
void VideoRxLoop(DtDevice& Dta2110)
{
    // Initialize an RxFifo object and associate it with the first (and only) network 
    // port of the DTA-2110.
    // Note: A hardware pipe is needed for optimal video reception performance. However,
    // if it's unavailable (maybe in use by another program), no exception will be raised
    // at this point.
    // Instead, the exception will be thrown when executing VideoRxFifo.Start().
    AvFifo::RxFifo VideoRxFifo{};
    VideoRxFifo.Attach(Dta2110, 1, AvFifo::HwOrSwPipe::ForceHwPipe);

    // Configure the IP and SMPTE 2110 receive parameters.
    VideoRxFifo.SetIpPars(AvFifo_VideoRx_IpPars);
    VideoRxFifo.Configure(AvFifo_VideoRx_Config);

    // Initialize the SDL context through its constructor.
    SdlContext Sdl{};

    // Initiate video frame reception. Two types of failures may occur:
    //  1. Exceptions: Unusual circumstances that hinder the proper execution of the 
    //     program. Examples include system issues like driver errors or unavailability 
    //     of a required hardware pipe.
    //  2. Errors: Problems from the operational environment, like a non-existent IP
    //     address. In such cases, RxFifo::Start() will return false.
    if (!VideoRxFifo.Start())
    {
        // If VideoRxFifo fails to start, retrieve the RxFifo error using GetStatus().
        AvFifo::FifoStatus Status{VideoRxFifo.GetStatus()};

        // Generate a human-readable error message for clarity and problem-solving.
        cout << "Failed to start: " << AvFifo::FifoStatusToMessage(Status) << '\n';

        // To prevent potential resource leakage, detach the VideoRxFifo.
        VideoRxFifo.Detach();

        // Halt further execution due to the failure in video frame reception.
        return;
    }

    // Main loop: continue until a quit event is detected.
    // In this example, the QuitFlag is only set by an SDL quit event.
    bool QuitFlag{false};
    while (!QuitFlag)
    {
        // Check FIFO status, on failure print message and break out of the loop.
        AvFifo::FifoStatus Status{VideoRxFifo.GetStatus()};
        if (Status != AvFifo::FifoStatus::Ok)
        {
            cout << "Failure: " << AvFifo::FifoStatusToMessage(Status) << '\n';
            break;
        }

        // Poll for SDL events. If an SDL 'quit' event is detected (closing the window), 
        // break out of the loop.
        SDL_Event SdlEvent;
        while (SDL_PollEvent(&SdlEvent))
            if (SdlEvent.type == SDL_QUIT)
                QuitFlag = true;

        if (!QuitFlag)
        {
            // If a video frame is available, render it in the SDL Window and return the
            // frame to the memory pool.
            if (VideoRxFifo.GetFifoLoad() > 0)
            {
                AvFifo::Frame* Frm{VideoRxFifo.Read()};
                Sdl.Render(Frm);
                VideoRxFifo.ReturnToMemPool(Frm);
            } else {
                // Sleep briefly if no frame is available and no errors are present.
                std::this_thread::sleep_for(10ms);
            }
        }
    }

    // Stop receiving frames and detach the RxFifo object as part of the cleanup process.
    VideoRxFifo.Stop();
    VideoRxFifo.Detach();
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- SdlContext::RenderFrame -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Renders an AvFifo frame to an SDL window.
//
void SdlContext::Render(AvFifo::Frame* Frm)
{
    SDL_UpdateTexture(Texture, nullptr, Frm->Data(), FrameWidth*2);
    SDL_RenderCopy(Renderer, Texture, nullptr, nullptr);
    SDL_RenderPresent(Renderer);
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- SdlContext::SdlContext -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Initializes SDL for video rendering, and create a window, texture, and renderer for
// displaying video in a window.
//
SdlContext::SdlContext()
{
    if (SDL_Init(SDL_INIT_VIDEO) < 0)
        throw runtime_error("SDL initialization failed: " + string(SDL_GetError()));

    Window = SDL_CreateWindow("AvFifo_VideoRx - Video Window",
                               SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, 
                               FrameWidth/Scale, FrameHeight/Scale, SDL_WINDOW_RESIZABLE);
    if (!Window)
        throw runtime_error("SDL_CreateWindow failed: " + string(SDL_GetError()));

    Renderer = SDL_CreateRenderer(Window, -1, 0);
    if (!Renderer)
        throw runtime_error("SDL_CreateRenderer failed: " + string(SDL_GetError()));

    // The texture uses the UYVY pixel format.
    Texture = SDL_CreateTexture(Renderer, SDL_PIXELFORMAT_UYVY,
                                    SDL_TEXTUREACCESS_STREAMING, FrameWidth, FrameHeight);
    if (!Texture)
        throw runtime_error("SDL_CreateTexture failed: " + string(SDL_GetError()));
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- SdlContext::~SdlContext -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Destroys the SDL resources. 
//
SdlContext::~SdlContext()
{
    if (Texture)
    {
        SDL_DestroyTexture(Texture); 
        Texture = nullptr;
    }
    if (Renderer) 
    {
        SDL_DestroyRenderer(Renderer); 
        Renderer = nullptr;
    }
    if (Window) 
    {
        SDL_DestroyWindow(Window);
        Window = nullptr;
    }
}
