// *#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* AvFifo_VideoTx.cpp *#*#*#*#*#*#*#*#*# (C) 2023 DekTec
//
// Code example AvFifo_VideoTx - SMPTE 2110 video generator
// 
// This code serves as a practical code example of generating video frames and using
// the TxFifo functionality to transmit a SMPTE 2110 video stream.
// The resultant output visualizes a simulation of moving and colliding balls.
// 
// The example includes comprehensive error handling similar to that in AvFifo_VideoRx. 
// However, the commentary in this file is more condensed. For a more detailed 
// explanation on error handling, please refer to the comments in AvFifo_VideoRx.
// 
// Note:
//  - This example uses a fixed IP address, port number, video format, etc.
//    Configuration parameters are defined in 'Configuration.h'.
//

const char* Version{"Jun2023.2"};

// -.-.-.-.-.-.-.-.-.-.-.- Include necessary headers and libraries -.-.-.-.-.-.-.-.-.-.-.-

#include "Configuration.h"          // Constants for configuring this specific example.
#include "DTAPI_AvFifo.h"           // DekTec's AvFifo API, part of DTAPI.
#include "Generator.h"              // SMPTE 2110 video frame generator.
#include <iostream>                 // Standard library: Console input/output operations.
#include <thread>                   // Standard library: Thread support.

using namespace std;                // Avoid std:: prefix for standard library calls.
using namespace AvFifo;             // Avoid AvFifo:: prefix, keeping code concise.

// Forward declarations to enable top-down function presentation.
void AvFifo_VideoTx();
void VideoTxLoop(DtDevice&);

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- main -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// The main function initiates the execution of the program and manages fatal exceptions.
//
int main(int argc, char* argv[])
{
    cout << "\n-----------------------------------------------------------------------\n"
         << "DekTec AvFifo API Example - AvFifo_VideoTx - SMPTE 2110 video generator\n"
         << "Version " << Version << "\n"
         << "-----------------------------------------------------------------------\n\n"
         << "DTA-2110 device number: " << Dta2110DeviceNo << "\n"
         << "Destination IP address: "
         << (int)AvFifo_VideoTx_IpPars.IpAddr[0] << "."
         << (int)AvFifo_VideoTx_IpPars.IpAddr[1] << "."
         << (int)AvFifo_VideoTx_IpPars.IpAddr[2] << "."
         << (int)AvFifo_VideoTx_IpPars.IpAddr[3] << ", "
         << "port " << AvFifo_VideoTx_IpPars.Port << "\n"
         << "Video format          : 10-bit packed, 4:2:2 UYVY chroma subsampling, "
         << FrameWidth << "x" << FrameHeight << "\n"
         << "Moving ball simulation: " << NumMovingBalls << " balls, "
         << "radius " << MinBallRadius << "-" << MaxBallRadius << ", "
         << "initial speed " << MinBallSpeed << "-" << MaxBallSpeed << "pixels/frame\n\n"
         << "Close this window to quit the application.\n";

    try {AvFifo_VideoTx();}
    catch (AvFifo::UsageError& e) {cerr << "Usage error: " << e.what() << '\n';}
    catch (exception& e) {cerr << "Exception: " << e.what() << '\n';}
    catch (...) {cerr << "Caught an unknown exception\n";}

    std::this_thread::sleep_for(3000ms);
    return 0;
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- AvFifo_VideoTx -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Attaches to the hardware, invokes the main loop, and detaches from the hardware.
//
void AvFifo_VideoTx()
{
    // Attach DtDevice object to the DTA-2110 PCIe card with index 'Dta2110DeviceNo'.
    DtDevice Dta2110{};
    DTAPI_RESULT Res{Dta2110.AttachToType(2110, Dta2110DeviceNo)};
    if (Res != DTAPI_OK)
        throw runtime_error("DTA-2110 attach fails:" + string(DtapiResult2Str(Res)));

    // Validate network card operation before starting as it is a prerequisite for 
    // successful execution.
    Res = Dta2110.IsNetworkCardOperational(1,
                        AvFifo_VideoTx_IpPars.IpVersion==AvFifo::IpProtocolVersion::IPv4,
                        AvFifo_VideoTx_IpPars.IpVersion==AvFifo::IpProtocolVersion::IPv6);
    if (Res != DTAPI_OK)
    {
        Dta2110.Detach();
        throw runtime_error("Networking error: " + string(DtapiResult2Str(Res)));
    }

    // Network card is operational, proceed to invoke the main loop.
    try {VideoTxLoop(Dta2110);}
    catch (AvFifo::DriverError& e) {cerr << "Driver error: " << e.what() << '\n';}
    catch (AvFifo::HardwarePipeUnavailable& e) {cerr << e.what() << '\n';}
    catch (AvFifo::SchedulingError& e) {cerr << e.what() << '\n';}

    // Cleanup: Detach the Dta2110 object from the hardware.
    Dta2110.Detach();
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- VideoTxLoop -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// VideoTxLoop demonstrates the use of the AvFifo API for video transmission using a 
// DTA-2110 device. It sets up an AvFifo TxFifo object for video transmission, configures
// IP and SMPTE 2110 parameters, and calculates the frame size.
//
// It uses a frame generator to create a dynamic video feed simulating moving colored 
// balls. The function continually generates and sends frames within a main loop, 
// given enough FIFO space and until an exit signal is received.
//
// The function concludes by stopping transmission and detaching the TxFifo from the
// device, ensuring a clean shutdown.
//
void VideoTxLoop(DtDevice& Dta2110)
{
    // Initialize an TxFifo object and associate it with the first (and only) network 
    // port of the DTA-2110. Request a hardware pipe as we want to transmit video.
    AvFifo::TxFifo VideoTxFifo{};
    VideoTxFifo.Attach(Dta2110, 1, AvFifo::HwOrSwPipe::ForceHwPipe);

    // Configure the IP and SMPTE 2110 transmit parameters.
    VideoTxFifo.SetIpPars(AvFifo_VideoTx_IpPars);
    VideoTxFifo.Configure(AvFifo_VideoTx_Config);

    // Compute the linear frame size (#bytes in the frame).
    size_t FrameSize{FrameWidth*FrameHeight*20 / 8};

    // Create and initialize the frame generator.
    Generator FrameGen{};
    FrameGen.Initialize({FrameWidth, FrameHeight});

    // Initialize ToD to current time. 
    // Add 100ms to have time to create the first few frames.
    DtTimeOfDay ToD{};
    Dta2110.GetTimeOfDay(ToD);
    ToD += 100'000'000;             // 100,000,000ns = 100ms

    // Initiates video frame transmission. In case of failure, an exception is thrown 
    // or 'Start()' returns false.
    if (!VideoTxFifo.Start())
    {
        AvFifo::FifoStatus Status{VideoTxFifo.GetStatus()};
        cout << "Failed to start: " << AvFifo::FifoStatusToMessage(Status) << '\n';
        VideoTxFifo.Detach();
        return;
    }

    // Enter the main loop, which continues operation until a quit event is signaled. 
    // In this example, the QuitFlag is only set on failures.
    bool QuitFlag{false};
    while (!QuitFlag)
    {
        // Check FIFO status, on failure print message and set quit flag.
        AvFifo::FifoStatus Status{VideoTxFifo.GetStatus()};
        if (Status != AvFifo::FifoStatus::Ok)
        {
            cout << "Failure: " << AvFifo::FifoStatusToMessage(Status) << '\n';
            QuitFlag = true;
        }

        // Only create and write a new Frame to the TxFifo if it has space, otherwise 
        // we sleep.
        else if (VideoTxFifo.GetFifoLoad() < VideoTxFifo.GetMaxSize())
        {
            // Align frame transmission to the video media grid. 
            ToD = AvFifo::Tod2Grid_Video(ToD, {50, 1});

		    // Obtain a Frame from the memory pool, and assemble a video frame in it.
            AvFifo::Frame* Frame{VideoTxFifo.GetFrameFromMemPool(FrameSize)};
            Frame->ToD = ToD;
            Frame->RtpTime = AvFifo::St2110::Tod2Rtp_Video(ToD);
            FrameGen.Generate(Frame->Data());
            Frame->NumValidBytes = (int)FrameSize;

		    // Write the newly assembled Frame to the TxFifo for transmission.
            VideoTxFifo.Write(Frame);

		    // Advance the time-of-day (ToD) to the next frame time, which is 20ms later.
            ToD += 20'000'000;      // 20,000,000ns = 20ms
        } else {
            std::this_thread::sleep_for(10ms);
        }
    }

    // Cleanup: Stop transmitting frames and detach the TxFifo object.
    VideoTxFifo.Stop();
    VideoTxFifo.Detach();
}
